import datetime as dt
import unittest

import numpy as np
import pandas as pd
from numpy.testing import assert_allclose

from darkgreybox.fit import darkgreyfit
from darkgreybox.models import Ti


def error_metric(y: np.ndarray, Z: np.ndarray) -> float:
    return np.sum(y - Z)


class DarkGreyFitTest(unittest.TestCase):

    def test__darkgreyfit__returns_correct_dataframe__when_prefit_splits_specified(self):

        train_start = dt.datetime(2021, 1, 1, 1, 0)
        train_end = dt.datetime(2021, 1, 1, 6, 0)
        test_start = dt.datetime(2021, 1, 1, 7, 0)
        test_end = dt.datetime(2021, 1, 1, 9, 0)

        rec_duration = 1

        params = {
            'Ti0': {'value': 10, 'vary': False},
            'Ria': {'value': 1},
            'Ci': {'value': 1},
        }

        y_train = pd.Series([10, 10, 20, 20, 20, 30])

        X_train = pd.DataFrame(
            index=pd.date_range(train_start, train_end, freq=f'{rec_duration}H'),
            data={
                'Ta': [10, 10, 10, 20, 20, 20],
                'Ph': [0, 10, 0, 0, 10, 0],
                'Ti0': [10, 10, 20, 20, 20, 30]
            })

        X_test = pd.DataFrame(
            index=pd.date_range(test_start, test_end, freq=f'{rec_duration}H'),
            data={
                'Ta': [30, 30, 30],
                'Ph': [0, 10, 0],
            })

        y_test = pd.Series([30, 30, 40])

        models = [Ti(params, rec_duration)]

        ic_params_map = {
            'Ti0': lambda X_test, y_test, train_result: y_test.iloc[0],
        }

        prefit_splits = [([], [0, 1, 2]), ([], [3, 4, 5])]

        expected_columns = pd.MultiIndex.from_tuples([
            ('train', 'start_date'),
            ('train', 'end_date'),
            ('train', 'model'),
            ('train', 'model_result'),
            ('train', 'time'),
            ('train', 'method'),
            ('train', 'error'),
            ('test', 'start_date'),
            ('test', 'end_date'),
            ('test', 'model'),
            ('test', 'model_result'),
            ('test', 'time'),
            ('test', 'error')
        ])

        actual_df = darkgreyfit(
            models=models,
            X_train=X_train,
            y_train=y_train,
            X_test=X_test,
            y_test=y_test,
            ic_params_map=ic_params_map,
            error_metric=error_metric,
            prefit_splits=prefit_splits,
            reduce_train_results=True,
        )

        # assert on the returned dataframe being sane

        self.assertTrue(expected_columns.equals(actual_df.columns))

        self.assertListEqual([train_start], actual_df[('train', 'start_date')].tolist())
        self.assertListEqual([train_end], actual_df[('train', 'end_date')].tolist())
        self.assertListEqual([test_start], actual_df[('test', 'start_date')].tolist())
        self.assertListEqual([test_end], actual_df[('test', 'end_date')].tolist())

        self.assertIsInstance(actual_df[('train', 'model')].iloc[0], Ti)
        self.assertIsInstance(actual_df[('test', 'model')].iloc[0], Ti)

        assert_allclose(y_train.values, actual_df[('train', 'model_result')].iloc[0].Z, atol=0.01)
        assert_allclose(y_test.values, actual_df[('test', 'model_result')].iloc[0].Z, atol=0.01)



import copy
from timeit import default_timer as timer
from typing import Callable, Dict, List, Optional, cast

import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from darkgreybox import logger
from darkgreybox.base_model import DarkGreyModel


def train_models(
    models: List[DarkGreyModel],
    X_train: pd.DataFrame,
    y_train: pd.Series,
    error_metric: Callable,
    splits: Optional[List] = None,
    method: str = 'nelder',
    obj_func: Optional[Callable] = None,
    reduce_train_results: bool = False,
    n_jobs: int = -1,
    verbose: int = 10
) -> pd.DataFrame:
    """
    Trains the `models` for the given `X_train` and `y_train` training data
    for `splits` using `method`.

    Params:
        models: list of `model.DarkGreyModel` objects
            list of models to be trained
        X_train: `pandas.DataFrame`
            A pandas DataFrame of the training input data X
        y_train: `pandas.Series`
            A pandas Series of the training input data y
        error_metric: Callable
            An error metric function that confirms to the `sklearn.metrics` interface
        splits: Optional[List]
            A list of training data indices specifying sub-sections of `X_train` and `y_train`
            for the models to be trained on
        method : str
            Name of the fitting method to use. Valid values are described in:
            `lmfit.minimize`
        obj_func: Optional[Callable]
            The objective function to minimise during the fitting
        n_jobs: Optional[int]
            The number of parallel jobs to be run as described by `joblib.Parallel`
        verbose: Optional[int]
            The degree of verbosity as described by `joblib.Parallel`

    Returns:
        `pd.DataFrame` with a record for each model's result for each split

    Example:
    ~~~~

    from sklearn.metrics import mean_squared_error
    from sklearn.model_selection import KFold

    from darkgreybox.model import TiTe
    from darkgreybox.fit import train_models


    prefit_df = train_models(
        models=[TiTe(train_params, rec_duration=1)],
        X_train=X_train,
        y_train=y_train,
        splits=KFold(n_splits=int(len(X_train) / 24), shuffle=False).split(X_train),
        error_metric=mean_squared_error,
        method='nelder',
        n_jobs=-1,
        verbose=10
    )
    ~~~~
    """

    logger.info('Training models...')

    if n_jobs != 1:
        with Parallel(n_jobs=n_jobs, verbose=verbose) as p:
            df = cast(pd.DataFrame, pd.concat(
                cast(pd. DataFrame, p(delayed(train_model)(
                    model, X_train.iloc[idx], y_train.iloc[idx], error_metric, method, obj_func)
                    for _, idx in splits or [(None, range(len(X_train)))] for model in models
                )),
                ignore_index=True
            ))

    else:
        df = cast(pd.DataFrame, pd.concat([
            train_model(model, cast(pd.DataFrame, X_train.iloc[idx]), y_train.iloc[idx], error_metric, method, obj_func)
            for _, idx in splits or [(None, range(len(X_train)))] for model in models
        ],
            ignore_index=True))

    if reduce_train_results:
        return reduce_results_df(df)
    else:
        return df


def train_model(
    base_model: DarkGreyModel,
    X_train: pd.DataFrame,
    y_train: pd.Series,
    error_metric: Callable,
    method: str = 'nelder',
    obj_func: Optional[Callable] = None,
) -> pd.DataFrame:
    """
    Trains a copy of `base_model` for the given `X_train` and `y_train` training data
    using `method`.

    Params:
        base_model: `model.DarkGreyModel`
            model to be trained (a copy will be made)
        X_train: `pandas.DataFrame`
            A pandas DataFrame of the training input data X
        y_train: `pandas.Series`
            A pandas Series of the training input data y
        error_metric: Callable
            An error metric function that confirms to the `sklearn.metrics` interface
        method : str
            Name of the fitting method to use. Valid values are described in:
            `lmfit.minimize`
        obj_func: Optional[Callable]
            The objective function to minimise during the fitting

    Returns:
        `pd.DataFrame` with a single record for the fit model's result
    """

    start = timer()
    model = copy.deepcopy(base_model)

    start_date = X_train.index[0]
    end_date = X_train.index[-1]

    X = X_train.to_dict(orient='list')
    y = cast(np.ndarray, y_train.values)

    try:
        model.fit(
            X,
            y,
            method=method,
            ic_params=get_ic_params(model, X_train),
            obj_func=obj_func
        )

    except ValueError:
        end = timer()
        return pd.DataFrame({
            'start_date': [start_date],
            'end_date': [end_date],
            'model': [np.NaN],
            'model_result': [np.NaN],
            'time': [end - start],
            'method': [method],
            'error': [np.NaN]
        })

    model_result = model.predict(X)
    end = timer()

    return pd.DataFrame({
        'start_date': [start_date],
        'end_date': [end_date],
        'model': [model],
        'model_result': [model_result],
        'time': [end - start],
        'method': [method],
        'error': [error_metric(y, model_result.Z)]
    })


def get_ic_params(model: DarkGreyModel, X_train: pd.DataFrame) -> Dict:
    """
    Returns the initial condition parameters of a model from the training data

    Params:
        model: `model.DarkGreyModel`
            model to get initial condition parameters from
        X_train: `pandas.DataFrame`
            A pandas DataFrame of the training input data X

    Returns:
        A dictionary containing the initial conditions and their corresponding values
        as defined by the training data

    """

    # TODO: this is horrible - make this clearer and more robust
    ic_params = {}
    for key in model.params:
        if '0' in key:
            if key in X_train:
                ic_params[key] = X_train.iloc[0][key]
            else:
                raise KeyError(f'Initial condition key {key} does not have corresponding X_train field')

    return ic_params


def reduce_results_df(df: pd.DataFrame, decimals: int = 6) -> pd.DataFrame:
    """
    Reduces `df` dataframe by removing nan and duplicate records

    Params:
        df: `pd.DataFrame`
            The dataframe to be reduced / cleaned
        decimal: int
            The number of decimal points for the float comparison when removing duplicates

    Returns :
        the reduced / cleaned `pd.DataFrame`
    """

    return (
        df.replace([-np.inf, np.inf], np.nan)
        .dropna()
        .round({'error': decimals})
        .sort_values('time')
        .drop_duplicates(subset=['error'], keep='first')
        .sort_values('error')
        .reset_index(drop=True)
    )



from abc import ABC
from copy import deepcopy
from typing import Callable, Dict, Optional, Union

from lmfit import Parameters, minimize
from lmfit.minimizer import MinimizerResult
import numpy as np

from darkgreybox import logger


class DarkGreyModelResult:
    '''
    Container object that holds the results of the model
    '''

    def __init__(self, Z, **kwargs):
        '''
        Parameters
        ----------
        Z : np.array
            The measured variable's fit / predicted values
        kwargs: kwargs
            Any other parameters calculated can be passed here.

        ~~~~
        E.g. in case of a TiTeTh model, Ti is the measured variable and Ti, Te, Th
        variables can be passed as kwargs for easy access

        DarkGreyModelResult(Ti, Ti=Ti, Te=Te, Th=Th)
        ~~~~
        '''

        self.Z = Z
        for key, val in kwargs.items():
            setattr(self, key, val)


class DarkGreyModel(ABC):
    '''
    Abstract Base Class for DarkGrey Models
    '''

    def __init__(self, params: Union[Parameters, Dict], rec_duration: float):
        '''
        Initialises the model instance

        Parameters
        ----------
        params : dict
            A dictionary of parameters for the fitting. Key - value pairs should follow the
            `lmfit.Parameters` declaration:
            e.g. {'A' : {'value': 10, 'min': 0, 'max': 30}} - sets the initial value and the bounds
            for parameter `A`
        rec_duration : float
            The duration of each measurement record in hours
        '''

        self.result = MinimizerResult()

        # convert the params dict into lmfit parameters
        if isinstance(params, Parameters):
            self.params = deepcopy(params)
        else:
            self.params = Parameters()
            for k, v in params.items():
                self.params.add(k, **v)

        # set the number of records based on the measured variable's values
        self.rec_duration = rec_duration

    def fit(
        self,
        X: Dict,
        y: np.ndarray,
        method: str,
        ic_params: Optional[Dict] = None,
        obj_func: Optional[Callable] = None
    ):
        '''
        Fits the model by minimising the objective function value

        Parameters
        ----------
        X : dict
            A dictionary of input values for the fitting - these values are fixed during the fit.
        y : np.array
            The measured variable's values for the minimiser to fit to
        method : str
            Name of the fitting method to use. Valid values are described in:
            `lmfit.minimize`
        ic_params : Optional[Dict]
            The initial condition parameters - if passed in these will overwrite
            the initial conditions in self.params
        obj_func : Optional[Callable]
            The objective function that is passed to `lmfit.minimize`/
            It must have (params, *args, **kwargs) as its method signature.
            Default: `def_obj_func`

        Returns
        -------
        `lmfit.MinimizerResult`
            Object containing the optimized parameters and several
            goodness-of-fit statistics.
        '''

        # overwrite initial conditions
        if ic_params is not None:
            for k, v in ic_params.items():
                if k in self.params:
                    self.params[k].value = v
                else:
                    logger.warning(f'Key `{k}` not found in initial conditions params')

        # we are passing X, y to minimise as kwargs
        self.result = minimize(
            obj_func or self.def_obj_func,
            self.params,
            kws={'model': self.model, 'X': X, 'y': y},
            method=method
        )

        self.params = self.result.params

        return self

    def predict(self, X: Dict, ic_params: Optional[Dict] = None) -> DarkGreyModelResult:
        '''
        Generates a prediction based on the result parameters and X.

        Parameters
        ----------
        X : dict
            A dictionary of input values
        ic_params : dict
            The initial condition parameters - if passed in these will overwrite
            the initial conditions in self.params

        Returns
        -------
        The results of the model
        '''

        if ic_params is not None:
            for k, v in ic_params.items():
                self.params[k].value = v

        return self.model(self.params, X)

    def model(self, params: Parameters, X: Dict) -> DarkGreyModelResult:
        '''
        A system of differential equations describing the thermal model
        '''
        ...  # pragma: no cover

    def lock(self):
        '''
        Locks the parameters by setting `vary` to False
        '''

        for param in self.params.keys():
            self.params[param].vary = False

        return self

    @staticmethod
    def def_obj_func(params: Parameters, *args, **kwargs):
        '''
        Default objective function
        Computes the residual between measured data and fitted data
        The model, X and y are passed in as kwargs by `lmfit.minimize`
        '''
        return ((kwargs['model'](params=params, X=kwargs['X']).Z - kwargs['y'])).ravel()
